/*************************************************************************\
*   				Xcard 1.2				  *
*									  *
* Copyright 1991, Kenneth C. Nelson			 	          *
*									  *
* Rights       : I, Ken Nelson own Xcard.  You may compile, modify, and   *
*                use Xcard.  You may not sell Xcard or anything derived   *
*		 from Xcard in binary or source for more than reasonable  *
*                media costs.                                             *
*                                                                         *
*		 Xcard uses a dynamic object library	   	          *
*		 that I don't own.  See the subdirectory dynobj for  	  *
*		 restrictions on its use.				  *
*								          *
*                Please feel free to modify Xcard. See Todo for details.  *
*									  *
\*************************************************************************/



/*
 * cards.c	- provides routines to make and use a cardfile.
 *
 */


#include <stdio.h>
#include <string.h>

#include "dyn.h"

#include "cards.h"

extern char *strdup();
extern char *strtrim();



/*
 * Define some strings used to block out parts of the physical card file.
 *
 */

#define VALID_XCARD_FILE "VALID_XCARD\n"

#define REC_MARK "&&##$$\n"


/*
 * Define what a cardfile is.
 *
 */

typedef struct Cardfile
{
  int		is_open;
  char		*title;
  char		*about;
  char		*template;
  char		*name;
  int		currentCard;
  DynObject	cards;	/* A dynamic array that will hold cards */
} Cardfile;



/*
 * The cardfile for the whole shebang!
 *
 */

static Cardfile theCardfile = {FALSE,NULL,NULL,NULL,NULL,0,NULL};



/*
 * cardFileIsOpen() - returns TRUE if the cardfile is open.
 *		      else returns FALSE.
 */

int cardFileIsOpen()
{
  return theCardfile.is_open;
}



/*
 * getCard() - local routine that gets a card. Returns NULL if not
 *	       found.
 *
 */
 
static Card *getCard(num)
  int num;
{
  Card *card;

 
   if (theCardfile.cards == NULL)
   {
     card = NULL;
   }
   else
   {
     card = (Card *) DynGet(theCardfile.cards,num);
   }
   return card;
  
}


/*
 * clearCardFile() - emptys the current CardFile.
 *
 */

void	clearCardFile()
{
  /*
   * Clear up the strings.
   *
   */
   
  if (theCardfile.about != NULL)
  {
    cfree(theCardfile.about);
  }

  if (theCardfile.template != NULL)
  {
    cfree(theCardfile.template);
  }

  if (theCardfile.name != NULL)
  {
    cfree(theCardfile.name);
  }
  theCardfile.name = NULL;
  
  theCardfile.currentCard = 0;
  if (theCardfile.cards != NULL)
  {
     DynDestroy(theCardfile.cards);
     theCardfile.cards = NULL;
  }

}





/*
 * saveCardFile() - saves the current cardfile in the same file as it
 *		    came from.
 */
 
int	saveCardFile()
{
  FILE *file;
  int  saveCurrent;
  char *body;
  char *title;
  char *template;
  char *about;
  
  if (cardFileIsOpen())
  {

     file = fopen(cardFileName(),"w");

     if (file != NULL)
     {
         fprintf(file,"%s",VALID_XCARD_FILE);

	 title = cardFileTitle();
	 if (title[strlen(title)-1] != '\n')
           fprintf(file,"%s\n",cardFileTitle());
	 else
           fprintf(file,"%s",cardFileTitle());

	 template = cardFileTemplate();
	 if (template[strlen(template)-1] != '\n')
  	   fprintf(file,"%s\n%s",cardFileTemplate(),REC_MARK);
	 else
  	   fprintf(file,"%s%s",cardFileTemplate(),REC_MARK);

	 about = aboutCardFile();
	 if (about[strlen(about)-1] != '\n')
	    fprintf(file,"%s\n%s",aboutCardFile(),REC_MARK);
	 else
	    fprintf(file,"%s%s",aboutCardFile(),REC_MARK);

	 
         saveCurrent = theCardfile.currentCard;

	 if (setCurrentCard(FIRSTCARD))
	 {
	   /*
	    * Write out each card.
	    *
	    */

	   do
	   {
	     fprintf(file,"%s\n",headerCurrentCard());
	     
	     body = bodyCurrentCard();
	     if (body[strlen(body)-1] != '\n')
  	       fprintf(file,"%s\n%s",bodyCurrentCard(),REC_MARK);
	     else
	       fprintf(file,"%s%s",bodyCurrentCard(),REC_MARK);
	     
	   } while (setCurrentCard(NEXTCARD));	   

	 }

	 setCurrentCard(saveCurrent);

	 fclose(file);
	 return TRUE;
     }
     else
     {
       printf("Could not open for output card file: %s\n",cardFileName());    
       return FALSE;
     }  
   }

}



/*
 * readRecord() -  reads a card from the current spot in the file until the stop
 *		   characters, building a string along the way.
 *		   Returns a newly allocated string.
 *
 */

static char *readRecord(file)
  FILE *file;
{
  char returnedLine[MAX_CARD_SIZE+1];
  char line[MAX_CARD_LINE+1];
  int  use_strcpy = TRUE;
  int keep_going = TRUE;

  strcpy(returnedLine,"");
  keep_going = (fgets(line,MAX_CARD_LINE,file) && strcmp(line,REC_MARK) != 0);
  while (!feof(file) && keep_going)
  {
      if (use_strcpy)
      {
        strcpy(returnedLine,line);
        use_strcpy = FALSE;
      }
      else
      {
        strcat(returnedLine,line);
      }
      keep_going = (fgets(line,MAX_CARD_LINE,file) && strcmp(line,REC_MARK) != 0);
  }
  
  return strdup(returnedLine);
  
}



/*
 * loadCardFile()  - first clears the current Cardfile if a newfile is true,
 *		     and then loads a new one from the file specified. 
 *
 */

int	loadCardFile(filename,newfile)
  char *filename;
  int  newfile;
{
  FILE *file  = NULL;
  int  status=FALSE;
  char line[MAX_CARD_LINE+1];
  char *card;
  char *header;
  char *body;
  
  file = fopen(filename,"r");
  if (file == NULL)
  {
    printf("error: could not open cardfile %s\n",filename);
    status = TRUE;
    if (newfile)
    {
      clearCardFile();
      theCardfile.name = filename;
    }
  }
  else
  {

    
    fgets(line,MAX_CARD_LINE,file);
    if (strcmp(line,VALID_XCARD_FILE) == 0)
    {

      if (newfile == NEW_FILE)
      {
	clearCardFile();
      }

	theCardfile.name = filename;

      /*
       * Now get the title.
       *
       */

       fgets(line,MAX_CARD_LINE,file);
       setCardFileTitle(line);

       /*
        * Read in the template and about records.
	*
	*/

	setCardFileTemplate(readRecord(file));
	setCardFileAbout(readRecord(file));
	
        while (!feof(file))
        {
	   card = readRecord(file);
	   if (strlen(card) > 0)
	   {
	     /*
	      * Now it gets tricky.  Find the end of the first line.
	      * Replace the carriage return with the end of a string.
	      * The original string now holds two strings, card and first line +1
	      * The first string is the header, the second the body.
	      */

	     body   = strchr(card,'\n');
	     *body  = '\0';
	     body   = (body+1);
	     header = card;

	     addCard(APPEND,header,body);
	     
	   }

        } /* not eof */

        fclose(file);
        status = TRUE;

    }
    else  /* invalid file */
    {
      printf("error: Invalid XCard file %s\n",filename);
      status = FALSE;
    }
  }

  theCardfile.is_open = status;
  return status; 
}





/*
 * sortCardFile() - sorts the card file into ASCENDING or DESCENDING order.
 *		    What a kludgy bubble sort!!
 *
 *
 */
 
int	sortCardFile(order)
  int order;
{
   int i, j;

   Card *iCard, *jCard, *tempCard;

   tempCard = (Card *) calloc(1,sizeof(Card));
   
   for (i=0; i < numCardsInCardFile()-1; i++)
   {
     iCard = (Card * )getCard(i);
     for (j=i+1; j < numCardsInCardFile(); j++)
     {
       jCard = (Card * )getCard(j);
       if (order == ASCENDING)
       {
         if (strcmp(iCard->header,jCard->header) > 0)
	 {
	    *tempCard = *iCard;
	    *iCard = *jCard;
	    *jCard = *tempCard;
	    
	 }
       }
       else
       {
         
         if (strcmp(iCard->header,jCard->header) < 0)
	 {
	    *tempCard = *jCard;
	    *jCard = *iCard;
	    *iCard = *tempCard;
	 }
       }       
     }
   }
}



/*
 * cardFileName() - returns the name of this cardfile.
 *
 */
 
char	*cardFileName()
{
  if (theCardfile.name != NULL)
  {
    return theCardfile.name;
  }
  else
  {
    return EMPTY_STRING;
  }
}



/*
 * cardFileTitle() - returns the name of this cardfile.
 *
 */
 
char	*cardFileTitle()
{
  if (theCardfile.title != NULL)
  {
    return theCardfile.title;
  }
  else
  {
    return EMPTY_STRING;
  }
}



/*
 * setCardFileTitle() - sets the title of this card file.
 *
 */
 
void setCardFileTitle(title)
 char *title;
{
  if (theCardfile.title != NULL)
  {
    cfree(theCardfile.title);
  }
  theCardfile.title = strdup(title);
}



/*
 * cardFileTemplate() - returns the template string for this card file.
 *			The template string is the string inserted automatically
 *			into new cards.
 */
 
char *cardFileTemplate()
{
  if (theCardfile.template == NULL)
  {
    return EMPTY_STRING;
  }
  else
  {
    return theCardfile.template;
  }
}



/*
 * setCardFileTemplate() - sets the template card.
 *
 */

void setCardFileTemplate(to_string)
  char *to_string;
{
  /*
   * Clean up some space.
   */
   
  if (theCardfile.template == NULL)
  {
    cfree(theCardfile.template);
    theCardfile.template = NULL;
  }

  /*
   * Copy it in.
   *
   */
   
  if (to_string != NULL)
  {
    theCardfile.template = strdup(to_string);
  }
}




/*
 * aboutCardFile() - returns the string which tells "about" this cardfile.
 *		     The user edits this when the cardfile is created.
 *
 */

char	*aboutCardFile()
{
 if (theCardfile.about == NULL)
 {
   return EMPTY_STRING;
 }
 else
 {
   return theCardfile.about;
 }
}





/*
 * setCardFileAbout() - sets the about string for the cardfile.
 *
 */

void setCardFileAbout(to_string)
  char *to_string;
{
  /*
   * Clean up some space.
   */
   
  if (theCardfile.about == NULL)
  {
    cfree(theCardfile.about);
    theCardfile.about = NULL;
  }

  /*
   * Copy it in.
   *
   */
   
  if (to_string != NULL)
  {
    theCardfile.about = strdup(to_string);
  }
}




/*
 * numCardInCardFile() - returns the number of cards in the card file.
 *
 */


int numCardsInCardFile()
{
  int size;
  
  if (theCardfile.cards !=  NULL)
  {
    size = DynSize(theCardfile.cards);
  }
  else
  {
   size = 0;
  }
  return size;
}


/*
 * currentCard() -  returns the position of the current card.
 *
 */

int currentCard()
{
  return theCardfile.currentCard;
}


/*
 * setCurrentCard() - sets the position of the current Card(). Valid
 *		      values are NEXTCARD,PREVCARD,FIRSTCARD,LASTCARD, or
 *		      a number greater than 0 which identifies a card
 *		      by number.  If the command is invalid or the position
 *		      would be invalid, FALSE is returned, else TRUE.
 *
 */
 
int	setCurrentCard(position)
{
  int status = FALSE;

  if (numCardsInCardFile() > 0)
  {
    switch (position)
    {
      case NEXTCARD:
		    if (theCardfile.currentCard + 1 < numCardsInCardFile())
		    {
		      theCardfile.currentCard++;
		      status = TRUE;
		    }
		    break;

      case PREVCARD:
		    if (theCardfile.currentCard > 0)
		    {
		      theCardfile.currentCard--;
		      status = TRUE;
		    }
		    break;

      case FIRSTCARD:
		     theCardfile.currentCard = DynLow(theCardfile.cards);
		     status = TRUE;
		    break;

      case LASTCARD:
		    theCardfile.currentCard = DynHigh(theCardfile.cards);
		    status = TRUE;
	           break;

      default:
		   if (position >= 0 && position <= numCardsInCardFile())
		   {
		     theCardfile.currentCard = position;
		     status = TRUE;
		   }
    }
    
  }

  return status;

}








void setHeaderCurrentCard(text)
  char *text;
{
  Card *theCard;

  theCard = getCard(theCardfile.currentCard);
  if (theCard != NULL)
  {
    if (theCard->header != NULL)
    {
      cfree(theCard->header);
    }
    theCard->header = strdup(text);
  }
}



void setBodyCurrentCard(text)
  char *text;
{
  Card *theCard;

  theCard = getCard(theCardfile.currentCard);
  if (theCard != NULL)
  {
    if (theCard->body != NULL)
    {
      cfree(theCard->body);
    }
    theCard->body = strdup(text);
  }
}


/*
 * headerCurrentCard() - returns the current card header.
 *
 */
 
char     *headerCurrentCard()
{

   Card *theCard;
   
   theCard = getCard(currentCard());
   if (theCard != NULL)
   {
     return theCard->header;
   }
   else
   {
     return EMPTY_STRING;
   }
}


/*
 * bodyCurrentCard() - returns the body of the current card.
 *
 */
 

char	*bodyCurrentCard()
{
   Card *theCard;
   
   theCard = getCard(currentCard());
   if (theCard != NULL)
   {
     return theCard->body;
   }
   else
   {
     return EMPTY_STRING;
   }
}


/*
 * rmCurrentCard() - deletes the current card. 
 *
 */

int	rmCurrentCard()
{
  if (theCardfile.currentCard >= 0)
  {
    DynDelete(theCardfile.cards,theCardfile.currentCard);
  }
}



/*
 * addCard() - adds a Card after the current card.
 *
 */

int	addCard(where,header,body)
  int  where;  /* APPEND or INSERT */
  char *header;
  char *body;
{
   Card *newCard = (Card *) calloc(1,sizeof(Card));

   newCard->deleted= FALSE;
   newCard->header = strdup(header);
   newCard->body   = strdup(body);   

   /*
    * If this is the first card, then create the array
    * to hold them. The dynamic array is re-allocated
    * every 100 inserts.
   */
   
   if (theCardfile.cards == NULL)
   {
     theCardfile.cards = DynCreate(sizeof(Card),100);
   }
   if (where == INSERT)
     DynInsert(theCardfile.cards,currentCard(),newCard);
   else
     DynAdd(theCardfile.cards,newCard);

   setCurrentCard(NEXTCARD);
}


